#include "stdafx.h"
#include "Allocator.h"

///////////////////////////////////////////////////////////////////////////////

MemMediaSample::MemMediaSample()
	: mRefCount(0)
	, mpAllocator(NULL)
	, mpBufferBase(NULL)
{
}

MemMediaSample::~MemMediaSample() {
	if (mpBufferBase)
		HeapFree(GetProcessHeap(), 0, mpBufferBase);
}

HRESULT MemMediaSample::Init(IMemAllocator *pAlloc, const ALLOCATOR_PROPERTIES& props) {
	long size = props.cbPrefix + props.cbBuffer + props.cbAlign - 1;
	mpBufferBase = HeapAlloc(GetProcessHeap(), 0, size);
	if (!mpBufferBase)
		return E_OUTOFMEMORY;

	mpBuffer = (char *)(((size_t)mpBufferBase + props.cbAlign - 1) & ~((size_t)props.cbAlign - 1)) + props.cbPrefix;
	mpAllocator = pAlloc;
	return S_OK;
}

ULONG STDMETHODCALLTYPE MemMediaSample::AddRef() {
	return (ULONG)InterlockedIncrement(&mRefCount);
}

ULONG STDMETHODCALLTYPE MemMediaSample::Release() {
	LONG rv = InterlockedDecrement(&mRefCount);

	if (!rv)
		mpAllocator->ReleaseBuffer(this);

	return rv;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::QueryInterface(REFIID riid, void **ppvObject) {
	if (riid == IID_IUnknown) {
		*ppvObject = static_cast<IUnknown *>(this);
		AddRef();
		return S_OK;
	}

	if (riid == IID_IMediaSample) {
		*ppvObject = static_cast<IMediaSample *>(this);
		AddRef();
		return S_OK;
	}

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::GetPointer(BYTE **ppBuffer) {
	*ppBuffer = (BYTE *)mpBuffer;
	return S_OK;
}

long STDMETHODCALLTYPE MemMediaSample::GetSize() {
	return mBufferSize;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::GetTime(REFERENCE_TIME *pTimeStart, REFERENCE_TIME *pTimeEnd) {
	if (!mbStreamStartTimeSet)
		return VFW_E_SAMPLE_TIME_NOT_SET;

	*pTimeStart	= mStreamTimeStart;
	*pTimeEnd	= mStreamTimeEnd;		// must occur even if we have no stop time

	if (!mbStreamStopTimeSet)
		return VFW_S_NO_STOP_TIME;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::SetTime(REFERENCE_TIME *pTimeStart, REFERENCE_TIME *pTimeEnd) {
	if (!pTimeStart) {
		mbStreamStartTimeSet = false;
		mbStreamStopTimeSet = false;
	} else {
		mbStreamStartTimeSet = true;
		mStreamTimeStart	= *pTimeStart;

		if (pTimeEnd) {
			mbStreamStopTimeSet = true;
			mStreamTimeEnd		= *pTimeEnd;
		} else {
			mbStreamStopTimeSet = false;
			mStreamTimeEnd		= mStreamTimeStart + 1;
		}
	}

	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::IsSyncPoint() {
	return mbIsSyncPoint ? S_OK : S_FALSE;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::SetSyncPoint(BOOL bIsSyncPoint) {
	mbIsSyncPoint = (bIsSyncPoint != 0);
	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::IsPreroll() {
	return mbIsPreroll ? S_OK : S_FALSE;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::SetPreroll(BOOL bIsPreroll) {
	mbIsPreroll = (bIsPreroll != 0);
	return S_OK;
}

long STDMETHODCALLTYPE MemMediaSample::GetActualDataLength() {
	return mDataLength;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::SetActualDataLength(long Length) {
	if (Length > mBufferSize || Length < 0)
		return VFW_E_BUFFER_OVERFLOW;

	mDataLength = Length;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::GetMediaType(AM_MEDIA_TYPE **ppMediaType) {
	MediaTypePtr p(mpMediaType);

	if (!p && mpMediaType) {
		*ppMediaType = NULL;
		return E_OUTOFMEMORY;
	}

	*ppMediaType = p.Detach();
	return *ppMediaType ? S_OK : S_FALSE;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::SetMediaType(AM_MEDIA_TYPE *pMediaType) {
	mpMediaType = pMediaType;
	if (pMediaType && !mpMediaType)
		return E_OUTOFMEMORY;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::IsDiscontinuity() {
	return mbIsDiscontinuity ? S_OK : S_FALSE;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::SetDiscontinuity(BOOL bDiscontinuity) {
	mbIsDiscontinuity = (bDiscontinuity != 0);
	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::GetMediaTime(LONGLONG *pTimeStart, LONGLONG *pTimeEnd) {
	if (!mbMediaTimesSet)
		return VFW_E_MEDIA_TIME_NOT_SET;

	*pTimeStart	= mMediaTimeStart;
	*pTimeEnd	= mMediaTimeEnd;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemMediaSample::SetMediaTime(LONGLONG *pTimeStart, LONGLONG *pTimeEnd) {
	if (!pTimeStart || !pTimeEnd) {
		mbMediaTimesSet = false;
	} else {
		mbMediaTimesSet = true;
		mMediaTimeStart	= *pTimeStart;
		mMediaTimeEnd	= *pTimeEnd;
	}

	return S_OK;
}

///////////////////////////////////////////////////////////////////////////////

MemAllocator::MemAllocator()
	: mRefCount(0)
	, mBufferAvailable(NULL)
	, mProps()
	, mBuffersOutstanding(0)
	, mWaiters(0)
{
}

MemAllocator::~MemAllocator() {
	if (mBufferAvailable) {
		// free currently available buffers and prevent any new ones from going
		// out
		Decommit();

		mLock.Enter();
		while(mBuffersOutstanding) {
			mLock.Exit();
			WaitForSingleObject(mBufferAvailable, INFINITE);
			mLock.Enter();
		}
		mLock.Exit();

		// free any buffers left
		Decommit();

		CloseHandle(mBufferAvailable);
	}
}

HRESULT MemAllocator::Init() {
	mBufferAvailable = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (!mBufferAvailable)
		return E_OUTOFMEMORY;

	return S_OK;
}

ULONG STDMETHODCALLTYPE MemAllocator::AddRef() {
	return (ULONG)InterlockedIncrement(&mRefCount);
}

ULONG STDMETHODCALLTYPE MemAllocator::Release() {
	LONG rv = InterlockedDecrement(&mRefCount);

	if (!rv)
		delete this;

	return rv;
}

HRESULT STDMETHODCALLTYPE MemAllocator::QueryInterface(REFIID riid, void **ppvObject) {
	if (riid == IID_IMemAllocator) {
		*ppvObject = static_cast<IMemAllocator *>(this);
		AddRef();
		return S_OK;
	}

	if (riid == IID_IUnknown) {
		*ppvObject = static_cast<IUnknown *>(this);
		AddRef();
		return S_OK;
	}

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

HRESULT STDMETHODCALLTYPE MemAllocator::SetProperties(ALLOCATOR_PROPERTIES *pRequest, ALLOCATOR_PROPERTIES *pActual) {
	if (!pRequest)
		return E_POINTER;

	mLock.Enter();
	bool active = mSamples.empty();
	mLock.Exit();

	if (active)
		return VFW_E_ALREADY_COMMITTED;

	if (pRequest->cbAlign <= 0)
		return VFW_E_BADALIGN;

	if (pRequest->cbAlign & (pRequest->cbAlign - 1))
		return VFW_E_BADALIGN;

	mProps = *pRequest;

	if (mProps.cBuffers < 1)
		mProps.cBuffers = 1;

	if (mProps.cbBuffer < 1)
		mProps.cbBuffer = 1;

	if (mProps.cbPrefix < 0)
		mProps.cbPrefix = 0;

	if (pActual)
		*pActual = mProps;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemAllocator::GetProperties(ALLOCATOR_PROPERTIES *pProps) {
	*pProps = mProps;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemAllocator::Commit() {
	if (!mProps.cbBuffer)
		return VFW_E_SIZENOTSET;

	mLock.Enter();
	bool alreadyCommitted = mbCommitted;
	bool buffersOutstanding = mBuffersOutstanding > 0;
	mLock.Exit();

	if (alreadyCommitted)
		return VFW_E_ALREADY_COMMITTED;

	if (buffersOutstanding)
		return VFW_E_BUFFERS_OUTSTANDING;

	// free any existing buffers
	Decommit();

	std::vector<MemMediaSample *> samples;
	samples.reserve(mProps.cBuffers);

	long i;
	for(i=0; i<mProps.cBuffers; ++i) {
		MemMediaSample *samp = new(std::nothrow) MemMediaSample;
		if (!samp)
			break;

		HRESULT hr = samp->Init(this, mProps);
		if (FAILED(hr)) {
			delete samp;
			break;
		}
	}

	if (i >= mProps.cBuffers)
		return S_OK;

	while(!samples.empty()) {
		delete samples.back();
		samples.pop_back();
	}

	return E_OUTOFMEMORY;
}

HRESULT STDMETHODCALLTYPE MemAllocator::Decommit() {
	std::vector<MemMediaSample *> samples;

	mLock.Enter();
	mbCommitted = false;
	samples.swap(mSamples);
	int waiters = mWaiters;
	mLock.Exit();

	while(waiters)
		SetEvent(mBufferAvailable);

	while(!samples.empty()) {
		delete samples.back();
		samples.pop_back();
	}

	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemAllocator::GetBuffer(IMediaSample **ppBuffer, REFERENCE_TIME *pStartTime, REFERENCE_TIME *pEndTime, DWORD dwFlags) {
	HRESULT hr;
	IMediaSample *pBuffer = NULL;

	mLock.Enter();
	for(;;) {
		if (!mbCommitted) {
			hr = VFW_E_NOT_COMMITTED;
			break;
		}

		if (!mSamples.empty()) {
			pBuffer = mSamples.back();
			mSamples.pop_back();
			++mBuffersOutstanding;
			hr = S_OK;
			break;
		}

		if (dwFlags & AM_GBF_NOWAIT) {
			hr = VFW_E_TIMEOUT;
			break;
		}

		++mWaiters;
		mLock.Exit();
		DWORD rv = WaitForSingleObject(mBufferAvailable, INFINITE);
		mLock.Enter();
		--mWaiters;

		if (rv != WAIT_OBJECT_0) {
			hr = E_FAIL;
			break;
		}
	}
	mLock.Exit();

	*ppBuffer = pBuffer;
	if (pBuffer) {
		pBuffer->AddRef();
		pBuffer->SetTime(pStartTime, pEndTime);
		pBuffer->SetSyncPoint(!(dwFlags & AM_GBF_NOTASYNCPOINT));
		pBuffer->SetDiscontinuity(0 != (dwFlags & AM_GBF_PREVFRAMESKIPPED));
	}
	return hr;
}

HRESULT STDMETHODCALLTYPE MemAllocator::ReleaseBuffer(IMediaSample *pBuffer) {
	mLock.Enter();
	mSamples.push_back(static_cast<MemMediaSample *>(pBuffer));
	--mBuffersOutstanding;
	mLock.Exit();

	SetEvent(mBufferAvailable);

	return S_OK;
}
