#include "stdafx.h"
#include "BasePin.h"
#include "BaseFilter.h"

///////////////////////////////////////////////////////////////////////////////

class EmptyMediaTypeEnumerator : public IEnumMediaTypes {
	EmptyMediaTypeEnumerator& operator=(const EmptyMediaTypeEnumerator&);
	EmptyMediaTypeEnumerator(const EmptyMediaTypeEnumerator&);
public:
	EmptyMediaTypeEnumerator();
	~EmptyMediaTypeEnumerator();

	ULONG STDMETHODCALLTYPE AddRef();
	ULONG STDMETHODCALLTYPE Release();
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppvObject);

	HRESULT STDMETHODCALLTYPE Next(ULONG cMediaTypes, AM_MEDIA_TYPE **ppMediaTypes, ULONG *pcFetched);
	HRESULT STDMETHODCALLTYPE Skip(ULONG cMediaTypes);
	HRESULT STDMETHODCALLTYPE Reset();
	HRESULT STDMETHODCALLTYPE Clone(IEnumMediaTypes **ppEnum);

protected:
	LONG mRefCount;
};

EmptyMediaTypeEnumerator::EmptyMediaTypeEnumerator()
	: mRefCount(0)
{
}

EmptyMediaTypeEnumerator::~EmptyMediaTypeEnumerator() {
}

ULONG STDMETHODCALLTYPE EmptyMediaTypeEnumerator::AddRef() {
	return (ULONG)InterlockedIncrement(&mRefCount);
}

ULONG STDMETHODCALLTYPE EmptyMediaTypeEnumerator::Release() {
	LONG rv = InterlockedDecrement(&mRefCount);

	if (!rv)
		delete this;

	return (ULONG)rv;
}

HRESULT STDMETHODCALLTYPE EmptyMediaTypeEnumerator::QueryInterface(REFIID riid, void **ppvObject) {
	if (riid == IID_IEnumMediaTypes) {
		*ppvObject = static_cast<IEnumMediaTypes *>(this);
		AddRef();
		return S_OK;
	}

	if (riid == IID_IUnknown) {
		*ppvObject = static_cast<IUnknown *>(this);
		AddRef();
		return S_OK;
	}

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

HRESULT STDMETHODCALLTYPE EmptyMediaTypeEnumerator::Next(ULONG cMediaTypes, AM_MEDIA_TYPE **ppMediaTypes, ULONG *pcFetched) {
	if (pcFetched)
		*pcFetched = 0;

	return S_FALSE;
}

HRESULT STDMETHODCALLTYPE EmptyMediaTypeEnumerator::Skip(ULONG cMediaTypes) {
	return S_FALSE;
}

HRESULT STDMETHODCALLTYPE EmptyMediaTypeEnumerator::Reset() {
	return S_OK;
}

HRESULT STDMETHODCALLTYPE EmptyMediaTypeEnumerator::Clone(IEnumMediaTypes **ppEnum) {
	if (!ppEnum)
		return E_POINTER;

	*ppEnum = new(std::nothrow) EmptyMediaTypeEnumerator;
	if (!*ppEnum)
		return E_OUTOFMEMORY;

	(*ppEnum)->AddRef();
	return S_OK;
}

///////////////////////////////////////////////////////////////////////////////

BasePin::BasePin()
	: mpId(NULL)
	, mpConnectedPin(NULL)
	, mpFilter(NULL)
{
}

BasePin::~BasePin() {
}

void BasePin::SetPinInfo(BaseFilter *pFilter, LPCWSTR id) {
	mpId = id;
	mpFilter = pFilter;
}

HRESULT BasePin::Activate() {
	return S_OK;
}

void BasePin::Deactivate() {
}

ULONG STDMETHODCALLTYPE BasePin::AddRef() {
	return mpFilter->AddRef();
}

ULONG STDMETHODCALLTYPE BasePin::Release() {
	return mpFilter->Release();
}

HRESULT STDMETHODCALLTYPE BasePin::QueryInterface(REFIID riid, void **ppvObject) {
	if (riid == IID_IPin) {
		*ppvObject = static_cast<IPin *>(this);
		AddRef();
		return S_OK;
	}

	if (riid == IID_IUnknown) {
		*ppvObject = static_cast<IUnknown *>(this);
		AddRef();
		return S_OK;
	}

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

HRESULT STDMETHODCALLTYPE BasePin::Disconnect() {
	if (!mpFilter->IsStopped())
		return VFW_E_NOT_STOPPED;

	if (mpConnectedPin) {
		mpConnectedPin->Release();
		mpConnectedPin = NULL;
	}

	mpConnectedMediaType = NULL;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE BasePin::ConnectedTo(IPin **pPin) {
	if (!pPin)
		return E_POINTER;

	*pPin = mpConnectedPin;
	if (!mpConnectedPin)
		return VFW_E_NOT_CONNECTED;

	mpConnectedPin->AddRef();
	return S_OK;
}

HRESULT STDMETHODCALLTYPE BasePin::ConnectionMediaType(AM_MEDIA_TYPE *pmt) {
	if (!pmt)
		return E_POINTER;

	if (!mpConnectedMediaType)
		return VFW_E_NOT_CONNECTED;

	void *pbFormat = NULL;
	if (mpConnectedMediaType->pbFormat) {
		pbFormat = CoTaskMemAlloc(mpConnectedMediaType->cbFormat);
		if (!pbFormat)
			return E_OUTOFMEMORY;
		memcpy(pbFormat, mpConnectedMediaType->pbFormat, mpConnectedMediaType->cbFormat);
	}

	*pmt = *mpConnectedMediaType;
	pmt->pbFormat = (BYTE *)pbFormat;
	if (pmt->pUnk)
		pmt->pUnk->AddRef();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE BasePin::QueryPinInfo(PIN_INFO *pInfo) {
	if (!pInfo)
		return E_POINTER;

	PIN_DIRECTION pindir;
	HRESULT hr = QueryDirection(&pindir);
	if (FAILED(hr))
		return hr;

	pInfo->pFilter = mpFilter;
	mpFilter->AddRef();

	pInfo->dir = pindir;
	wcsncpy(pInfo->achName, mpId, MAX_PIN_NAME - 1);
	pInfo->achName[MAX_PIN_NAME - 1] = 0;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE BasePin::QueryId(LPWSTR *Id) {
	if (!Id)
		return E_POINTER;

	*Id = SysAllocString(mpId);
	if (!*Id)
		return E_OUTOFMEMORY;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE BasePin::EnumMediaTypes(IEnumMediaTypes **ppEnum) {
	if (!ppEnum)
		return E_POINTER;

	*ppEnum = new(std::nothrow) EmptyMediaTypeEnumerator;
	if (!*ppEnum)
		return E_OUTOFMEMORY;
	(*ppEnum)->AddRef();
	return S_OK;
}

HRESULT STDMETHODCALLTYPE BasePin::QueryInternalConnections(IPin **apPin, ULONG *nPin) {
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE BasePin::NewSegment(REFERENCE_TIME tStart, REFERENCE_TIME tStop, double dRate) {
	return S_OK;
}

///////////////////////////////////////////////////////////////////////////////

InputPin::InputPin() {
}

InputPin::~InputPin() {
	if (mpConnectedPin)
		mpConnectedPin->Release();
}

HRESULT STDMETHODCALLTYPE InputPin::Connect(IPin *pReceivePin, const AM_MEDIA_TYPE *pmt) {
	return E_UNEXPECTED;
}

HRESULT STDMETHODCALLTYPE InputPin::ReceiveConnection(IPin *pConnector, const AM_MEDIA_TYPE *pmt) {
	if (!pmt || !pConnector)
		return E_POINTER;

	if (mpConnectedPin)
		return VFW_E_ALREADY_CONNECTED;

	HRESULT hr = QueryAccept(pmt);
	if (hr != S_OK)
		return VFW_E_TYPE_NOT_ACCEPTED;

	mpConnectedMediaType = pmt;
	if (!mpConnectedMediaType)
		return E_OUTOFMEMORY;

	mpConnectedPin = pConnector;
	pConnector->AddRef();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE InputPin::QueryAccept(const AM_MEDIA_TYPE *pmt) {
	return S_OK;
}

HRESULT STDMETHODCALLTYPE InputPin::BeginFlush() {
	return S_OK;
}

HRESULT STDMETHODCALLTYPE InputPin::EndFlush() {
	return S_OK;
}

HRESULT STDMETHODCALLTYPE InputPin::EndOfStream() {
	return S_OK;
}

HRESULT STDMETHODCALLTYPE InputPin::QueryDirection(PIN_DIRECTION *pPinDir) {
	*pPinDir = PINDIR_INPUT;
	return S_OK;
}

///////////////////////////////////////////////////////////////////////////////

OutputPin::OutputPin() {
}

OutputPin::~OutputPin() {
}

HRESULT STDMETHODCALLTYPE OutputPin::Connect(IPin *pReceivePin, const AM_MEDIA_TYPE *pmt) {
	return VFW_E_NO_TRANSPORT;
}

HRESULT STDMETHODCALLTYPE OutputPin::ReceiveConnection(IPin *pConnector, const AM_MEDIA_TYPE *pmt) {
	return E_UNEXPECTED;
}

HRESULT STDMETHODCALLTYPE OutputPin::QueryAccept(const AM_MEDIA_TYPE *pmt) {
	if (!mpConnectedMediaType)
		return S_FALSE;

	return DSAreMediaTypesEqual(mpConnectedMediaType, pmt) ? S_OK : S_FALSE;
}

HRESULT STDMETHODCALLTYPE OutputPin::BeginFlush() {
	return E_UNEXPECTED;
}

HRESULT STDMETHODCALLTYPE OutputPin::EndFlush() {
	return E_UNEXPECTED;
}

HRESULT STDMETHODCALLTYPE OutputPin::EndOfStream() {
	return E_UNEXPECTED;
}

HRESULT STDMETHODCALLTYPE OutputPin::QueryDirection(PIN_DIRECTION *pPinDir) {
	*pPinDir = PINDIR_OUTPUT;
	return S_OK;
}
