#include "stdafx.h"
#include "SampleGrabber.h"

//#define VDDEBUG_DSSG VDDEBUG
#define VDDEBUG_DSSG (void)sizeof printf

///////////////////////////////////////////////////////////////////////////////

DSSampleGrabberInputPin::DSSampleGrabberInputPin()
	: mMediaType(GUID_NULL)
	, mSubType(GUID_NULL)
	, mTargetTimeStart(0)
	, mTargetTimeEnd(0)
	, mTimeBase(0)
	, mLastSeenTime(0)
	, mLastSeenTimeSnapshot(0)
	, mpCB(NULL)
	, mbRunning(false)
	, mbBufferSamples(false)
	, mbSampleCaptured(false)
	, mbRestartOnNewSegment(false)
{
}

bool DSSampleGrabberInputPin::GetEstimatedStreamPosition(REFERENCE_TIME *pos) {
	bool success = false;

	mLock.Enter();
	if (mbBufferSamples) {
		success = true;
		*pos = mLastSeenTimeSnapshot;
	}
	mLock.Exit();

	return success;
}

void DSSampleGrabberInputPin::SetRunning(bool running) {
	mLock.Enter();
	mbRunning = running;
	if (!running)
		ClearQueueLocked();
	mLock.Exit();
}

bool DSSampleGrabberInputPin::CheckForSample(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend) {
	// Check if we have a sample that contains this time -- if so, process it immediately.
	// Discard all samples that are too old.
	mLock.Enter();

	IMediaSample *found = NULL;

	REFERENCE_TIME trelstart(tstart - mTimeBase);
	REFERENCE_TIME trelend(tend - mTimeBase);

	MediaSamples::iterator it(mMediaSamples.begin()), itEnd(mMediaSamples.end());
	while(it != itEnd) {
		const SampleEntry& samp = *it;

		// End time <= desired time? Discard.
		if (samp.mEndTime <= trelstart) {
			samp.mpSample->Release();

			it = mMediaSamples.erase(it);
			itEnd = mMediaSamples.end();
			continue;
		}

		// Contains our time? Take it.
		if (trelstart < samp.mEndTime && trelend > samp.mStartTime) {
			if (!found) {
				found = samp.mpSample;
				found->AddRef();
			}
		}

		++it;
	}

	mbSampleCaptured = true;
	bool success = false;
	if (found) {
		success = true;

		if (mpCB)
			mpCB->Process(found, mTimeBase);

		found->Release();
	}

	mLastSeenTimeSnapshot = mLastSeenTime;
	mLock.Exit();

	return success;
}

void DSSampleGrabberInputPin::ResetOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend) {
	mLock.Enter();
	mTargetTimeStart = tstart;
	mTargetTimeEnd = tend;
	mbSampleCaptured = true;
	mbBufferSamples = false;
	mbRestartOnNewSegment = true;
	mSignal.check();

	ClearQueueLocked();
	mLock.Exit();
}

bool DSSampleGrabberInputPin::ResumeOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend) {
	mLock.Enter();

	if (CheckForSample(tstart, tend)) {
		mLock.Exit();
		return true;
	}

	mTargetTimeStart = tstart;
	mTargetTimeEnd = tend;
	mbSampleCaptured = false;
	mbBufferSamples = true;
	mSignal.check();

	ClearQueueLocked();
	mLock.Exit();

	return false;
}

void DSSampleGrabberInputPin::ClearQueueLocked() {
	while(!mMediaSamples.empty()) {
		IMediaSample *samp = mMediaSamples.back().mpSample;
		mMediaSamples.pop_back();
		samp->Release();
	}
}

void DSSampleGrabberInputPin::SetCallback(IDSSampleGrabberCallback *pCB) {
	mpCB = pCB;
}

void DSSampleGrabberInputPin::SetMediaType(const GUID *type) {
	mMediaType = *type;
}

void DSSampleGrabberInputPin::SetMediaSubType(const GUID *subtype) {
	mSubType = *subtype;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::QueryAccept(const AM_MEDIA_TYPE *pmt) {
	if (mMediaType != GUID_NULL) {
		if (pmt->majortype != mMediaType)
			return S_FALSE;

		if (mSubType != GUID_NULL && pmt->subtype != mSubType)
			return S_FALSE;

		if (mMediaType == MEDIATYPE_Video) {
			if (pmt->formattype != FORMAT_VideoInfo && pmt->formattype != FORMAT_VideoInfo2)
				return S_FALSE;

			if (mSubType == GUID_NULL) {
				if (pmt->subtype != MEDIASUBTYPE_RGB24
					&& pmt->subtype != MEDIASUBTYPE_RGB32
					&& pmt->subtype != MEDIASUBTYPE_YUY2
					&& pmt->subtype != MEDIASUBTYPE_UYVY
					&& pmt->subtype != MEDIASUBTYPE_YV12
					&& pmt->subtype != MEDIASUBTYPE_YVU9
					&& pmt->subtype != MEDIASUBTYPE_RGB555
					&& pmt->subtype != MEDIASUBTYPE_RGB565)
					return S_FALSE;
			}
		} else if (mMediaType == MEDIATYPE_Audio) {
			if (pmt->formattype != FORMAT_WaveFormatEx)
				return S_FALSE;

			if (mSubType == GUID_NULL) {
				if (pmt->subtype != MEDIASUBTYPE_PCM
					&& pmt->subtype != MEDIASUBTYPE_PCMAudio_Obsolete)
					return S_FALSE;
			}
		}
	}

	return TransformInputPin::QueryAccept(pmt);
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::Receive(IMediaSample *pSample) {
	HRESULT hr = S_OK;

	mLock.Enter();
	if (mbFlushing || !mbRunning) {
		hr = S_FALSE;
	} else if (mbBufferSamples || !mbSampleCaptured) {
		REFERENCE_TIME t1, t2;
		hr = pSample->GetTime(&t1, &t2);

		bool valid = false;
		if (hr == S_OK) {
			valid = true;

			// reject if sample time bracket is prior to desired time
			REFERENCE_TIME trelstart = mTargetTimeStart - mTimeBase;
			REFERENCE_TIME trelend = mTargetTimeEnd - mTimeBase;
			if (t2 <= trelstart)
				valid = false;

			if (t1 >= trelend && !mbSampleCaptured) {
				VDDEBUG_DSSG("SampleGrabber[%ls]: Time missed -- feeding late sample ([%d,%d] after [%d,%d])\n", mpFilter->GetNameInGraph(), (int)t1, (int)t2, (int)trelstart, (int)trelend);
				t1 = trelstart;
			}

			VDDEBUG_DSSG("SampleGrabber[%ls]: Seen [%d, %d]\n", mpFilter->GetNameInGraph(), (int)(t1 + mTimeBase), (int)(t2 + mTimeBase));
			mLastSeenTime = t2 + mTimeBase;
		}

		if (valid) {
			if (!mbSampleCaptured) {
				bool cont = true;

				if (mpCB)
					cont = mpCB->Process(pSample, mTimeBase);

				if (!cont) {
					mSignal.signal();
					mbSampleCaptured = true;
				}
			}

			if (mbBufferSamples && mbSampleCaptured) {
				SampleEntry& ent = mMediaSamples.push_back();
				ent.mpSample = pSample;
				pSample->AddRef();

				ent.mStartTime = t1;
				ent.mEndTime = t2;
			}
		}
	}
	mLock.Exit();

	if (mpAssocOutputPin)
		mpAssocOutputPin->Output(pSample);

	return hr;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::BeginFlush() {
	HRESULT hr = TransformInputPin::BeginFlush();

	mLock.Enter();
	ClearQueueLocked();
	mLock.Exit();

	return hr;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::EndOfStream() {
	mLock.Enter();
	mbBufferSamples = false;
	if (!mbSampleCaptured) {
		mbSampleCaptured = true;
		// It may seem to do this under lock, but it's so that CheckForSample() can atomically
		// clear the signal state while changing state variables.
		mSignal.signal();
	}

	HRESULT hr = TransformInputPin::EndOfStream();
	mLock.Exit();
	return hr;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::NewSegment(REFERENCE_TIME tStart, REFERENCE_TIME tStop, double dRate) {
	mLock.Enter();
	mTimeBase = tStart;
	HRESULT hr = TransformInputPin::NewSegment(tStart, tStop, dRate);
	ClearQueueLocked();

	if (mbRestartOnNewSegment) {
		mbRestartOnNewSegment = false;
		mbSampleCaptured = false;
		mbBufferSamples = true;
	}
	mLock.Exit();

	return hr;
}

///////////////////////////////////////////////////////////////////////////////

DSSampleGrabber::DSSampleGrabber() {
	mInputPin.SetAssociatedOutput(&mOutputPin);
	mOutputPin.SetAssociatedInput(&mInputPin);

	ReservePins(2);
	AddPin(&mInputPin, L"in");
	AddPin(&mOutputPin, L"out");
}

DSSampleGrabber::~DSSampleGrabber() {
}

void DSSampleGrabber::SetCallback(IDSSampleGrabberCallback *pCB) {
	mInputPin.SetCallback(pCB);
}

void DSSampleGrabber::SetMediaType(const GUID *type) {
	mInputPin.SetMediaType(type);
}

void DSSampleGrabber::SetMediaSubType(const GUID *subtype) {
	mInputPin.SetMediaSubType(subtype);
}

bool DSSampleGrabber::GetEstimatedStreamPosition(REFERENCE_TIME *pos) {
	return mInputPin.GetEstimatedStreamPosition(pos);
}

bool DSSampleGrabber::CheckForSample(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend) {
	return mInputPin.CheckForSample(tstart, tend);
}

void DSSampleGrabber::ResetOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend) {
	mInputPin.ResetOneShot(tstart, tend);
}

bool DSSampleGrabber::ResumeOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend) {
	return mInputPin.ResumeOneShot(tstart, tend);
}

void DSSampleGrabber::WaitForSample() {
	HANDLE h = mInputPin.GetSignal().getHandle();

	for(;;) {
		DWORD rval = MsgWaitForMultipleObjects(1, &h, FALSE, INFINITE, QS_ALLEVENTS);

		if (rval != WAIT_OBJECT_0 + 1)
			break;

		MSG msg;
		while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);

			if (WaitForSingleObject(h, 0) == WAIT_OBJECT_0)
				return;
		}
	}
}

HRESULT STDMETHODCALLTYPE DSSampleGrabber::GetClassID(CLSID *pClassID) {
	// {F3557C75-0FE6-45fd-87D9-E1CA8D444C46}
	static const GUID CLSID_DSSampleGrabber = { 0xf3557c75, 0xfe6, 0x45fd, { 0x87, 0xd9, 0xe1, 0xca, 0x8d, 0x44, 0x4c, 0x46 } };

	*pClassID = CLSID_DSSampleGrabber;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabber::Stop() {
	mInputPin.SetRunning(false);

	return BaseFilter::Stop();
}

HRESULT STDMETHODCALLTYPE DSSampleGrabber::Pause() {
	mInputPin.SetRunning(true);

	return BaseFilter::Pause();
}

HRESULT STDMETHODCALLTYPE DSSampleGrabber::Run(REFERENCE_TIME tStart) {
	mInputPin.SetRunning(true);

	return BaseFilter::Run(tStart);
}
